
def parse_dict(args, attr, default_ok=True):
    '''Parse an "nargs='*'" field from an ArgumentParser's parsed arguments

    Each argument value should be of the form <k>=<v>, and a dict() of those mappings will be returned

    At most one argument value can simply be of the form <v>, in which case it will be assigned to the None value in the returned dict
    '''
    arg = getattr(args, attr)
    ret = {}
    if isinstance(arg, list):
        for entry in arg:
            pcs = entry.split('=', 1)
            if len(pcs) == 2:
                [ repository, arg ] = pcs
                ret[repository] = arg
            elif len(pcs) == 1 and default_ok:
                if None in ret:
                    raise ValueError(f'Invalid --{attr}: multiple default values ({ret[None]}, {pcs[0]})')
                ret[None] = arg
            else:
                raise ValueError(f'Invalid --{attr}: {arg}')
    return ret



from re import match

from ..process import *


def ls(): return lines('git','remote')


LS_REMOTE_LINE_REGEX = '(?P<sha>[0-9a-f]{40})\s+(?:refs/(?P<type>[^/]+)/(?P<name>.*)|(?P<head>HEAD))'
def parse_ls_remote_lines(lns, head=None, tag=None, sha=None, heads=False, tags=False):
    d = {}
    for ln in lns:
        m = match(LS_REMOTE_LINE_REGEX, ln)
        typ = m['type']
        if typ:
            name = m['name']
            if typ not in d:
                d[typ] = {}
            if not sha or m['sha'].startswith(sha):
                d[typ][name] = m['sha']
        elif m['head']:
            if not sha or m['sha'].startswith(sha):
                d['head'] = m['sha']
        else:
            raise ValueError(f'Unexpected `git ls-remote` line: {ln}')

    if head:
        return d.get('heads', {}).get(head)
    if tag:
        return d.get('tags', {}).get(tag)
    if sha:
        r = {}
        for k,v in d.items():
            if k == 'head':
                if v.startswith(sha):
                    r[k] = v
            else:
                typ = { name:id for name,id in v.items() if id.startswith(sha) }
                if typ:
                    r[k] = typ
        d = r
    if heads or tags:
        if heads and tags: return { 'heads': d.get('heads'), 'tags': d.get('tags') }
        elif heads: return d.get('heads')
        elif tags: return d.get('tags')
        else: raise

    return d


def ls_remote(remote, *args, head=None, tag=None, sha=None, heads=False, tags=False):
    cmd = ['git','ls-remote']
    if head or heads: cmd += ['--heads']
    if tag or tags: cmd += ['--tags']
    cmd += (remote,) + args
    lns = lines(cmd)
    return parse_ls_remote_lines(lns, head=head, tag=tag, sha=sha, heads=heads, tags=tags)


def exists(name): return name in ls()


def url(name, *args, **kwargs):
    if not exists(name):
        if len(args) == 1:
            default = args[0]
        elif 'default' in kwargs:
            default = kwargs.pop('default')
        else:
            raise ValueError(f"Remote {name} doesn't exist")

        return default
    
    return line('git','remote','get-url',name)
_url = url


def init(name, url, branch=None, remote_branch=None, fetch=True, checkout=True, push=True):
    if exists(name):
        existing_url = _url(name)
        if existing_url != url:
            run('git','remote','set-url',name,url)
    else:
        run('git','remote','add',name,url)
    
    if fetch:
        run('git','fetch',name)
    
    if branch:
        remote_branch = remote_branch or 'master'
        upstream = f'{name}/{remote_branch}'
        try:
            run('git','branch','-u',upstream,branch)
        except CalledProcessError as e:
            if push:
                print(f'Failed to track upstream branch {upstream}; attempting to push {branch}:{remote_branch}')
                run('git','push',name,f'{branch}:{remote_branch}')
                run('git','branch','-u',upstream,branch)
            else:
                raise e
        if checkout:
            run('git','checkout',branch)


def push(name=None, local=None, remote=None):
    refspec = None
    if local:
        if remote:
            refspec = f'{local}:{remote}'
        else:
            refspec = f'{local}:{local}'
    
    run('git','push',name,refspec)




import setuptools
from setuptools import find_packages

from utz import *


class Compute:
    '''Container for computing fallback values to pass to setup()'''

    def long_description(self):
        '''Read in README.md as the `long_description`'''
        if exists('README.md'):
            with open('README.md', "r") as fh:
                return fh.read()

    def long_description_content_type(self):
        '''Default to markdown format for `long_description`'''
        return "text/markdown"

    VERSION_TAG_REGEX = r'v?(?P<version>(?P<base>\d+\.\d+\.\d+)(?P<rc>(?:r|c|rc|a|b)\d+)?(?:-(?P<commits_ahead>\d+)-g(?P<sha>[0-9a-f]{6,}))?)'
    def version(self):
        '''Infer version from git tag of the form "v_._._" (which must be present)'''
        try:
            tag = line('git','describe','--tags','HEAD')
            m = match(self.VERSION_TAG_REGEX, tag)
            if not m:
                raise ValueError('Unrecognized tag: %s' % tag)
            version = m['version']
        except CalledProcessError:
            version = line('git','log','--format=%h','-n1')

        return version

    def name(self):
        '''`name` defaults to the name of the containing directory'''
        return basename(getcwd())

    def author(self):
        '''Infer author name from most recent commit'''
        return line('git','log','-n1','--format=%an')

    def author_email(self):
        '''Infer author email from most recent commit'''
        return line('git','log','-n1','--format=%ae')

    def description(self):
        '''Set `description` to the contents of a <p> first-child of an initial <h1>'''
        try:
            import mistune
        except ImportError:
            import pip
            pip.main(['install','mistune>=0.8.1,<2'])  # match nbconvert
            import mistune

        html = mistune.Markdown()(self.long_description())

        try:
            import lxml
        except ImportError:
            import pip
            pip.main(['install','lxml'])
            import lxml

        from lxml.html import fragments_fromstring
        [ h1, p, *_ ] = fragments_fromstring(html)
        if h1.tag != 'h1' or p.tag != 'p':
            raise ValueError('Expected initial <h1> followed by <p> while parsing `description` from README.md')

        return p.text_content()

    def packages(self):
        return find_packages()

    def classifiers(self):
        classifiers = [
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ]

        license = self.license()
        if license:
            classifiers += [f"License :: OSI Approved :: {license} License"]

        return classifiers

    def python_requires(self): return '>=3.6'

    def license(self):
        if exists('LICENSE'):
            with open('LICENSE','r') as f:
                lines = f.readlines()
                first_lines = [ line.strip() for line in lines[:2] ]
                if first_lines == ['Apache License','Version 2.0, January 2004',]: return 'Apache v2'
                if first_lines == ['MIT License','',]: return 'MIT'


def setup(**kwargs):
    c = Compute()
    def compute(*keys):
        for k in keys:
            if k not in kwargs:
                fn = getattr(c, k)
                v = fn()
                s = str(v)
                s = s[:100] + '…' if len(s) > 100 else s
                print('Fallback value: %s = %s' % (k, s))
                kwargs[k] = v

    compute(
        'name',
        'version',
        'description',
        'long_description',
        'long_description_content_type',
        'license',
        'author',
        'author_email',
        'packages',
        'classifiers',
        'python_requires',
    )

    setuptools.setup(**kwargs)




from contextlib import contextmanager
from os.path import basename
from sys import stderr
from typing import Iterable

import utz
from utz import check, git, line, match, now, print_exc, run, tmpdir, uuid1


@contextmanager
def tmp(
    url,
    *clone_args,
    branch=None,
    ref=None,
    submodules=True,
    push=False,
    pull=False,
    cd=True,
    name=None,
    bare=False,
    **run_kwargs,
):
    '''contextmanager for creating a Git repo in a temporary directory, and optionally cd'ing into it and upstreaming
    commits from it

    - `url`: local or remote path to Git repository to be cloned
    - `clone_args` (List[str]): passed directly to `git clone`
    - `branch`: Git branch to operate on inside temporary clone (by default, inherit from origin)
    - `ref`: if `branch` doesn't already exist, reset it to this ref. Note: if `ref` is omitted, `branch` must already
        exist in the remote! If `ref` is set and `branch` isn't, a temporary/nonced branch name will be used.
    - `submodules`: recursively clone submodules (`git clone --recurse-submodules`)
    - `push` (str | bool | List[str]): run a `git push` after `yield`ing; `str` or `List[str]` serve as arguments to
        `git push`
    - `pull` (bool): upstream changes post-`yield` by `cd`ing into the origin directory (must be a local dir, otherwise
        use `push`) and running a `git pull` from this temporary clone
    - `cd` (bool): move into the temporary clone dir before `yield`ing
    - `name`: basename for the temporary clone directory (defaults to basename of `url`)
    - `bare`: clone a bare repository
    '''

    name = name or basename(url)
    if name.endswith('.git'): name = name[:-len('.git')]
    with tmpdir(name) as repo_dir:
        cmd = ['git','clone']
        if submodules: cmd += ['--recurse-submodules']
        if bare: cmd += ['--bare']
        if branch and not ref:
            # branch must already exist and we want to clone and work on it
            cmd += ['-b',branch]
        cmd += clone_args
        cmd += [ url, repo_dir, ]
        run(*cmd, **run_kwargs)
        if ref:
            with utz.cd(repo_dir):
                if ref is True:
                    ref = git.branch.current()

                if not branch:
                    branch = f'tmp-{now("short")}'

                if git.branch.exists(branch):
                    make_branch = False
                else:
                    make_branch = True

                if make_branch:
                    if bare:
                        run('git','branch',branch,ref)
                        run('git','symbolic-ref','HEAD',f'refs/heads/{branch}')
                    else:
                        run('git','checkout','-b',branch,ref)

                if cd:
                    yield repo_dir
        if cd:
            if not ref:
                # if `checkout or init`, we've already cd'd and yielded above
                with utz.cd(repo_dir):
                    yield repo_dir
        else:
            yield repo_dir

        # optionally cd into upstream `url` dir (assumed to be a local directory if `pull` is set) and pull changes
        # from the temporary `repo_dir`
        if pull:
            with utz.cd(url):
                if not branch: branch = git.branch.current()
                run('git','fetch',repo_dir)
                try:
                    ref_line = line('git','ls-remote',repo_dir,f'refs/heads/{branch}')
                    m = match(
                        '(?P<sha>[0-9a-f]{40})\s+refs/heads/%s' % branch,
                        ref_line,
                    )
                    if m:
                        sha = m['sha']
                    else:
                        raise RuntimeError(f'Unrecognized ls-remote line: {ref_line}')
                except Exception as e:
                    stderr.write('%s\n' % f'Error fetching branch {branch} from remote {repo_dir}')
                    print_exc()
                    sha = git.sha('FETCH_HEAD')

                if not check('git','merge','-m','\n'.join([f'Merge {branch} from tmpdir','',repo_dir]),sha):
                    tmp_branch = f'conflict-{sha}'
                    run('git','branch',tmp_branch,sha)
                    stderr.write('Pulling from %s failed, aborting merge, saving to %s\n' % (repo_dir, tmp_branch))
                    run('git','merge','--abort')

        # optionally push back upstream
        if push:
            with utz.cd(repo_dir):
                cmd = ['git','push',]
                if isinstance(push, str):
                    cmd += ['origin',f'{push}:{push}',]
                elif isinstance(push, Iterable):
                    cmd += push
                elif push is True:
                    if branch:
                        cmd += ['origin',f'{branch}:{branch}',]
                run(*cmd)



#!/usr/bin/env python
# coding: utf-8

# In[ ]:


from dask import compute
import dask.dataframe as dd
import pandas as pd

def df_counts(df):
    '''Get notnull and unique counts for each column in a DataFrame'''
    dask = isinstance(df, dd.DataFrame)
    if dask:
        nuniques = compute(*[ df[k].nunique() for k in df ])
        col_counts = pd.DataFrame(
            [
                { 'col':col, 'nunique':nunique }
                for col, nunique
                in zip(df.columns, nuniques)
            ]
        ) \
        .set_index('col')

        notnulls = df.notnull().sum().compute().rename('notnull')

        return pd.concat([notnulls, col_counts], axis=1)
    else:
        col_counts = pd.DataFrame(
            [
                { 'col':k, 'nunique':df[k].nunique() }
                for k in df
            ]
        ) \
        .astype({'nunique':int}) \
        .set_index('col')

        notnulls = df.notnull().sum().rename('notnull')

        return pd.concat([notnulls, col_counts], axis=1)


# In[ ]:


def col_counts(df, *cols):
    counts = df.groupby(list(cols)).size()
    counts = counts[counts > 1]
    return counts.compute().sort_values()

